@@ -12,7 +12,7 @@ use Log::Log4perl qw(:easy);
 use Number::Bytes::Human qw( format_bytes parse_bytes );
 
 # ABSTRACT: Yet Another RESTful-Archive Service
-our $VERSION = '1.12'; # VERSION
+our $VERSION = '1.14'; # VERSION
 
 
 has secret => rand;
@@ -39,14 +39,14 @@ sub startup {
             $tx->req->content->on(body => sub {
                     my $content = shift;
                     my $md5_b64 = $content->headers->header('Content-MD5') or return;
-                    my $md5 = unpack 'H*', b($md5_b64)->b64_decode;
-                    my $disk = $tools->disk_for($md5) or return;
-                    my $tmpdir = join '/', $disk, 'tmp';
-                    -d $tmpdir or do { mkpath $tmpdir;  chmod 0777, $tmpdir; };
-                    -w $tmpdir or chmod 0777, $tmpdir;
                     $content->asset->on(
                         upgrade => sub {
                             #my ( $mem, $file ) = @_;
+                            my $md5 = unpack 'H*', b($md5_b64)->b64_decode;
+                            my $disk = $tools->disk_for($md5) or return;
+                            my $tmpdir = join '/', $disk, 'tmp';
+                            -d $tmpdir or do { mkpath $tmpdir;  chmod 0777, $tmpdir; };
+                            -w $tmpdir or chmod 0777, $tmpdir;
                             $_[1]->tmpdir($tmpdir);
                         }
                     );
@@ -201,7 +201,7 @@ Yars - Yet Another RESTful-Archive Service
 
 =head1 VERSION
 
-version 1.12
+version 1.14
 
 =head1 SYNOPSIS
 
@@ -261,9 +261,9 @@ requests of the form
  HEAD http://$host/file/$md5/$filename
  GET http://$host/bucket_map
 
-to store and retrieve files, where $host may be any of the
-hosts in the cluster, $md5 is the md5 of the content, and
-$filename is a filename for the content to be stored.  See
+to store and retrieve files, where C<$host> may be any of the
+hosts in the cluster, C<$md5> is the md5 of the content, and
+C<$filename> is a filename for the content to be stored.  See
 L<Yars::Routes> for documentation of other routes.
 
 Failover is handled in the following manner:
@@ -402,7 +402,7 @@ And from Perl:
  
  $y->download("test_file3", $md5);
 
-=head2 Multiple hosts
+=head2 Multiple servers
 
 =head3 set up the URL
 
@@ -548,6 +548,20 @@ and C<yarsclient> to use when the primary is not available:
  failover_urls:
    - http://yars1:9001
 
+=head3 randomizing the server choices
+
+In order to more evenly spread the load over each node in the Yars
+cluster, you can randomize the servers that the client considers the
+"primary" and the "failover(s)":
+
+ ---
+ # ~/etc/Yars.conf on client systems
+ % use List::Util qw( shuffle );
+ % my @url = shuffle map { "http://yars$_:9001" } 1..3;
+ url: <%= $url[0] %>
+ failover_urls:
+   - <%= $url[1] %>
+
 =head2 Accelerated downloads with nginx
 
 One of the advantages of Clustericious is that it integrates with
