@@ -1,7 +1,7 @@
 package Yars::Client;
 
 # ABSTRACT: Yet Another RESTful-Archive Service Client
-our $VERSION = '1.12'; # VERSION
+our $VERSION = '1.14'; # VERSION
 
 use strict;
 use warnings;
@@ -21,13 +21,14 @@ use Data::Dumper;
 use IO::Uncompress::Gunzip qw(gunzip $GunzipError);
 use Number::Bytes::Human qw( format_bytes parse_bytes );
 use File::Temp qw( tempdir );
+use File::HomeDir;
+use YAML::XS;
+use Mojo::JSON qw( encode_json );
 
-route_doc upload   => "<filename>";
+route_doc upload   => "<filename> [md5]";
 route_doc download => "<filename> <md5> [dir]";
 route_doc remove   => "<filename> <md5>";
 
-has bucket_map_cached  => sub { 0; }; # Computed on demand.
-
 route 'welcome'        => "GET",  '/';
 route 'bucket_map'     => "GET",  '/bucket_map';
 route 'disk_usage'     => "GET",  '/disk/usage';
@@ -72,6 +73,13 @@ sub new {
       # tx
       $_[1]->req->headers->header('X-Yars-Skip-Verify' => 'on');
       $_[1]->res->max_message_size(parse_bytes($self->_config->max_message_size_client(default => 53687091200)));
+      $_[1]->on(finish => sub {
+        my $tx = shift;
+        if(defined $tx->res->headers->header("X-Yars-Cache"))
+        {
+          $self->bucket_map_cached(0);
+        }
+      });
     });
     return $self;
 }
@@ -87,6 +95,61 @@ sub client {
     } : $self->SUPER::client;
 }
 
+sub bucket_map_cached {
+    my($self, $new) = @_;
+
+    state $fn = File::Spec->catfile(
+        File::HomeDir->my_dist_data('Yars', { create => 1 }),
+        'bucket_map_cache.yml',
+    );
+
+    if(defined $new) {
+        $self->{bucket_map_cached} = $new;
+
+        if(ref $new) {
+            YAML::XS::DumpFile($fn, $new);
+        } else {
+            unlink $fn;
+        }
+    }
+
+    elsif(! defined $self->{bucket_map_cached})
+    {
+        if(-r $fn)
+        {
+            my $cache = YAML::XS::LoadFile($fn);
+
+            # make sure the cache and config are in sync:
+            # ie, all of the urls in the config refer to
+            # hosts in the cache.
+            # NOTE: Considered doing this based on the
+            # md5 on the config, but:
+            #  - doing it on the Yars.conf file requires
+            #    changes to Clustericious::Config
+            #  - doing it on the $self->_config means that
+            #    you can't randomly choose a different
+            #    server in the mojo template of the config.
+            #    (see https://github.com/plicease/Yars#randomizing-the-server-choices)
+            my %r = map { $_ => 1 } values %$cache;
+            if(grep { ! $r{$_} } ($self->_config->url, $self->_config->failover_urls))
+            {
+                $self->{bucket_map_cached} = 0;
+                unlink $fn;
+            }
+            else
+            {
+                $self->{bucket_map_cached} = $cache;
+            }
+        }
+        else
+        {
+            $self->{bucket_map_cached} = 0;
+        }
+    }
+
+    $self->{bucket_map_cached};
+}
+
 sub _get_url {
 
     # Helper to create the Mojo URL objects
@@ -126,7 +189,7 @@ sub location {
 sub download {
     # Downloads a file and saves it to disk.
     my $self = shift;
-    my ( $filename, $md5, $dest_dir ) = @_;
+    my ( $filename, $md5, $dest ) = @_;
     my $abs_url;
     if (@_ == 1) {
         $abs_url = shift;
@@ -179,7 +242,21 @@ sub download {
         DEBUG "Received asset with size ".$res->content->asset->size;
         TRACE "Received headers : ".$res->headers->to_string;
 
-        my $out_file = $dest_dir ? $dest_dir . "/$filename" : $filename;
+        # host == 0 means we tried the assigned host
+        if($host == 0) {
+            my $prev = $res->previous;
+            # previous message in the chain was a 301
+            # Moved Permanently means our bucket map is
+            # wrong our out of date
+            if($prev && $prev->code == 301) {
+                $self->bucket_map_cached(0);
+            }
+        }
+
+        my $out_file = $dest
+            ? File::Spec->catfile((ref $dest eq 'SCALAR' ? tempdir(CLEANUP => 1) : $dest), $filename)
+            : $filename;
+
         DEBUG "Writing to $out_file";
         if (my $e = $res->headers->header("Content-Encoding")) {
             LOGDIE "unsupported encoding" unless $e eq 'gzip';
@@ -204,7 +281,18 @@ sub download {
             WARN "No md5 in response header";
             next;
         }
-        unless ($verify eq $md5) {
+        if ($verify eq $md5)
+        {
+            if(ref $dest eq 'SCALAR')
+            {
+                open my $fh, '<', $out_file;
+                binmode $fh;
+                $$dest = do { local $/; <$fh> };
+                close $fh;
+            }
+        }
+        else
+        {
             WARN "Bad md5 for file (got $verify instead of $md5)";
             WARN "Response headers : ".$res->headers->to_string;
             unlink $out_file or WARN "couldn't remove $out_file : $!";
@@ -287,20 +375,31 @@ sub _all_hosts {
 sub upload {
     my $self = shift;
     my $content = ref($_[-1]) eq 'SCALAR' ? pop : undef;
-    my $filename = pop;
-    my $nostash;
-    if (@_) {
-        # To avoid failover :
-        # yarsclient upload --nostash 1 foo
-        # $yc->upload("--nostash",1","foo")
-        # This is undocumented since it is only intended to be
-        # used on a server when balancing, not as a public interface.
 
-        if ($_[0] =~ /nostash$/) {
-            shift;
-            $nostash = shift;
-        }
+    my $nostash;
+    # OLD COMMENT (still valid):
+    #  To avoid failover:
+    #  yarsclient upload --nostash 1 foo
+    #  Yars::Client->new->upload("--nostash" => 1, 'foo');
+    #  This is undocumented since it is only intended to be
+    #  used on a server when blanacing, not as a public interface.
+    # UPDATE (Graham Ollis 2015-11-10)
+    #  This feature is only used by yars_fast_balance
+    #  (see Yars::Command::yars_fast_balance for implementation).
+    #  This used to match on /nostash$/ but the only place this
+    #  is used it specifies it as a command line option like thing
+    #  ("--nostash"), but that means that you can't upload files
+    #  that end in *nostash.  In truth the interface here should
+    #  be a little better thought out and a little less batshitcrazy.
+    if(defined $_[0] && $_[0] eq '--nostash')
+    {
+      shift;
+      $nostash = shift;
     }
+
+    my $filename = shift;
+    my $md5 = defined $_[0] && $_[0] =~ /^[0-9a-f]+$/i ? lc shift : undef;
+
     if (@_) {
         LOGDIE "unknown options to upload : @_";
     }
@@ -322,7 +421,7 @@ sub upload {
     # Don't read the file.
     my $basename = basename($filename);
     my $asset    = Mojo::Asset::File->new( path => $filename );
-    my $md5      = digest_file_hex($filename, 'MD5');
+    $md5 ||= digest_file_hex($filename, 'MD5');
 
     my @servers = $self->_all_hosts( $self->_server_for($md5) );
 
@@ -355,14 +454,14 @@ sub upload {
             if ($code) {
                 INFO "Failed to PUT to $host : $code $msg";
             } else {
-                INFO "PUT to $host failed : ".$msg;
+                INFO "PUT to $host failed : ". ref $msg ? encode_json($msg) : $msg;
             }
         }
     }
+    $self->res($tx->res);
     return '' if !$code || !$tx->res->is_status_class(200);
 
     DEBUG "Response : ".$tx->res->code." ".$tx->res->message;
-    $self->res($tx->res);
     return 'ok';
 }
 
@@ -454,7 +553,7 @@ Yars::Client - Yet Another RESTful-Archive Service Client
 
 =head1 VERSION
 
-version 1.12
+version 1.14
 
 =head1 SYNOPSIS
 
