@@ -1,7 +1,7 @@
 package Yars::Routes;
 
 # ABSTRACT: set up the routes for Yars.
-our $VERSION = '1.12'; # VERSION
+our $VERSION = '1.14'; # VERSION
 
 
 use strict;
@@ -59,6 +59,7 @@ sub _get {
         TRACE "$md5 should be on $url";
         # but check our local stash first, just in case.
         _get_from_local_stash($c,$filename,$md5) and return;
+        $c->res->headers->add("X-Yars-Cache" => 0);
         return $c->render_moved("$url/file/$md5/$filename");
     }
 
@@ -221,7 +222,7 @@ put '/file/#filename/:md5' => { md5 => 'calculate' } => sub {
 
     if ( $assigned_server ne $c->tools->server_url ) {
         TRACE "assigned $assigned_server != ".$c->tools->server_url;
-        return _proxy_to( $c, $assigned_server, $filename, $digest, $asset )
+        return _proxy_to( $c, $assigned_server, $filename, $digest, $asset, 0 )
               || _stash_locally( $c, $filename, $digest, $asset )
               || _stash_remotely( $c, $filename, $digest, $asset )
               || $c->render(status => 507, text => "Unable to proxy or stash");
@@ -311,7 +312,9 @@ sub _proxy_to {
    $tx->req->content->asset($asset);
    $tx = $c->tools->_ua->start($tx);
    if ($res = $tx->success) {
-       $c->res->headers->location($tx->res->headers->location);
+       my $headers = $c->res->headers;
+       $headers->location($tx->res->headers->location);
+       $headers->add("X-Yars-Cache" => 0) unless $temporary;
        $c->render(status => $tx->res->code, text => 'ok');
        return 1;
    }
@@ -659,7 +662,7 @@ Yars::Routes - set up the routes for Yars.
 
 =head1 VERSION
 
-version 1.12
+version 1.14
 
 =head1 SYNOPSIS
 
